{

This unit is a small graphics library for VGA adapters. It is primarily
an alternative to the VESA unit included with TMT. It supports various
modes for display purposes on a VGA or Super VGA monitor. The primary
mode is 320x200 by 256 colors. This is an excellent mode for games and
such. Note that no Assembler has been used in this module. It could
obviously be made faster by doing so, but it is done in Pascal for
compatability across multiple compilers (my requirement).

The routines contained in this module are as follows:

FUNCTION  GetVideoMode : Byte;

  This function returns the current video mode.

PROCEDURE SetVideoMode(desiredVideoMode : Byte);

  This procedure sets the video mode to the desired type. The
  available types are defined as constants for ease of use.

FUNCTION  GetPixel(pix2get_x, pix2get_y : Word) : Byte;

  This procedure gets a pixel from a given location X,Y.

PROCEDURE PutPixel(pix2set_x, pix2set_y : Word; pix2set_c : Byte);

  This procedure sets a given pixel at X,Y to a given color.

PROCEDURE Ellipse(exc, eyc, ea, eb : Integer; Color : Byte);

  This procedure draws an ellipse with a center at exc,eyc and a defined
  size of ea, eb in a given color.

PROCEDURE Line(lnx1, lny1, lnx2, lny2 : Integer; Color : Byte);

  This procedure draws a line from x1,y1 to x2,y2 in a given color.

PROCEDURE Circle(X,y,size: longint; Color: Byte);

  This procedure draws a circle at X,Y with a given radius of a certain
  color.

PROCEDURE HLine(X,y,x2: Integer; Color: Byte);

  This procedure draws a horizontal line from X,Y to X2 of a given
  color. It is faster than the Line procedure for horizontal lines.

PROCEDURE VLine(X,y,y2: Integer; Color: Byte);

  This procedure draws a vertical line from X,Y to Y2 of a given color.
  It is also faster than the Line procedure for vertical lines.

PROCEDURE DrawBar( X1,Y1,x2,y2 :Word; Color :Byte );

  This procedure draws a filled bar from X1,Y1 to X2,Y2 of a given
  color.

PROCEDURE Rectangle(X,y,xx,yy :Integer;Color :Byte);

  This procedure draws a rectangle from X,Y to XX,YY of a given color.
  This is not filled.

PROCEDURE GetPalette(index2get : Byte; VAR r_inte, g_inte, b_inte : Byte);

  This procedure gets the pallette of the current graphics mode. Refer
  to additional documentation on pallettes if you need more information.
  There is no requirement to use pallettes if you do not need special
  colors.

PROCEDURE SetPalette(index2set, r_inte, g_inte, b_inte : Byte);

  This procedure sets the pallette of the current graphics mode.

PROCEDURE BurstSetPalette(burstPalette : RGBPalette);

  This procedure sets the RGB pallette.

PROCEDURE WaitForRetrace;

  For older video adapters, this routine waits for retrace to avoid
  Snow.

PROCEDURE ClearScr;

  This routine clears the current graphics screen.

PROCEDURE screen_off;

  This routine turns off the video adapter. CAUTION should be used with
  this routine, as some video adapters do not take kindly to this.

PROCEDURE screen_on;

  This routine turns on the video adapter.

PROCEDURE msdelay(ms : Word);

  This routine delays a given amount of milliseconds.

PROCEDURE Ellipse2(exc, eyc, ea, eb, i, clr : Integer);

  This routine draws an ellipse as well from X,Y with size a,b and
  color.

PROCEDURE FillBox(X1,Y1,x2,y2:Integer; Col:Byte);

  This routine draws a filled box from X1,Y1 to X2,Y2 of color.

PROCEDURE polygon(X1,Y1,x2,y2,x3,y3,x4,y4:Integer; c:Byte);

  This routine draws a polygon with 4 points X1,Y1, X2,Y2, X3,Y3, X4,Y4
  of a given color c.

PROCEDURE WriteXY( X,y :Word; Color :Byte; Str :STRING );

  This routine writes a given string at X,Y of a given color in the
  current DOS Rom font.

}

UNIT _Graphic;

INTERFACE

CONST
   { Standard text }
   _40x25t        = $02;
   _80x25t        = $03;
   { Standard VGA }
   _640x480x2     = $11;
   _640x480x16    = $12;
   _320x200x256   = $13;

TYPE
  RGBPalette = ARRAY[0..767] OF Byte;


VAR
   GMaxX, GMaxY : Word;
   GMaxColor : Word;

IMPLEMENTATION

USES
   Dos;

FUNCTION GetVideoMode : Byte;
VAR
  tempVMode : Byte;
BEGIN
  ASM
    mov ah,$0f
    Int $10
    mov tempVMode,al
  END;
  GetVideoMode := tempVMode;
END;

PROCEDURE SetVideoMode(desiredVideoMode : Byte);
{ desiredVideoMode = $03 : 80x25 colour text
                     $13 : 320x200x256 monoplaned
                           video data from $A000:0000 to $A000:FFFF
}
BEGIN
  ASM
    mov ah,0
    mov al,desiredVideoMode;
    Int $10
  END;
  { Set some common variables for use in applications - only some modes }
  IF desiredVideoMode = _640x480x2 THEN BEGIN
     GMaxX     := 640;
     GMaxY     := 480;
     GMaxColor := 2;
  END;
  IF desiredVideoMode = _640x480x16 THEN BEGIN
     GMaxX     := 640;
     GMaxY     := 480;
     GMaxColor := 16;
  END;
  IF desiredVideoMode = _320x200x256 THEN BEGIN
     GMaxX     := 320;
     GMaxY     := 200;
     GMaxColor := 256;
  END;
END;

FUNCTION GetPixel(pix2get_x, pix2get_y : Word) : Byte;
BEGIN
  GetPixel := Mem[$A000 : pix2get_y * 320 + pix2get_x];
END;

PROCEDURE PutPixel(pix2set_x, pix2set_y : Word; pix2set_c : Byte);
BEGIN
  Mem[$A000 : pix2set_y * 320 + pix2set_x] := pix2set_c;
END;

PROCEDURE Ellipse(exc, eyc, ea, eb : Integer; Color : Byte);
VAR
  elx, ely : Integer;
  aa, aa2, bb, bb2, d, dx, dy : longint;
BEGIN
  elx := 0; ely := eb; aa := longint(ea) * ea; aa2 := 2 * aa;
  bb := longint(eb) * eb; bb2 := 2 * bb;
  d := bb - aa * eb + aa DIV 4; dx := 0; dy := aa2 * eb;
  PutPixel(exc, eyc - ely, Color); PutPixel(exc, eyc + ely, Color);
  PutPixel(exc - ea, eyc, Color); PutPixel(exc + ea, eyc, Color);

  WHILE (dx < dy) DO BEGIN
    IF (d > 0) THEN BEGIN Dec(ely); Dec(dy, aa2); Dec(d, dy); END;
    Inc(elx); Inc(dx, bb2); Inc(d, bb + dx);
    PutPixel(exc + elx, eyc + ely, Color);
    PutPixel(exc - elx, eyc + ely, Color);
    PutPixel(exc + elx, eyc - ely, Color);
    PutPixel(exc - elx, eyc - ely, Color);
  END;
  Inc(d, (3 * (aa - bb) DIV 2 - (dx + dy)) DIV 2);
  WHILE (ely > 0) DO BEGIN
    IF (d < 0) THEN BEGIN Inc(elx); Inc(dx, bb2); Inc(d, bb + dx); END;
    Dec(ely); Dec(dy, aa2); Inc(d, aa - dy);
    PutPixel(exc + elx, eyc + ely, Color);
    PutPixel(exc - elx, eyc + ely, Color);
    PutPixel(exc + elx, eyc - ely, Color);
    PutPixel(exc - elx, eyc - ely, Color);
  END;
END;

PROCEDURE Line(lnx1, lny1, lnx2, lny2 : Integer; Color : Byte);
VAR
  lndd, lndx, lndy, lnai, lnbi, lnxi, lnyi : Integer;
BEGIN
  IF (lnx1 < lnx2) THEN BEGIN lnxi := 1; lndx := lnx2 - lnx1;
  END ELSE BEGIN lnxi := (-1); lndx := lnx1 - lnx2; END;
  IF (lny1 < lny2) THEN BEGIN lnyi := 1; lndy := lny2 - lny1;
  END ELSE BEGIN lnyi := (-1); lndy := lny1 - lny2; END;
  PutPixel(lnx1, lny1, Color);
  IF (lndx > lndy) THEN BEGIN lnai := (lndy - lndx) * 2;
    lnbi := lndy * 2;
    lndd := lnbi - lndx;
    REPEAT IF (lndd >= 0) THEN BEGIN Inc(lny1, lnyi);
      Inc(lndd, lnai); END ELSE Inc(lndd, lnbi);
      Inc(lnx1, lnxi); PutPixel(lnx1, lny1, Color);
    UNTIL (lnx1 = lnx2);
  END ELSE BEGIN lnai := (lndx - lndy) * 2; lnbi := lndx * 2;
    lndd := lnbi - lndy;
    REPEAT IF (lndd >= 0) THEN BEGIN Inc(lnx1, lnxi);
      Inc(lndd, lnai); END ELSE Inc(lndd, lnbi);
      Inc(lny1, lnyi); PutPixel(lnx1, lny1, Color);
    UNTIL (lny1 = lny2);
  END;
END;

PROCEDURE GetPalette(index2get : Byte; VAR r_inte, g_inte, b_inte : Byte);
{ returns the r, g, and b values of a palette index }
BEGIN
  Port[$3C7] := index2get;
  r_inte := Port[$3C9];
  g_inte := Port[$3C9];
  b_inte := Port[$3C9];
END;

PROCEDURE SetPalette(index2set, r_inte, g_inte, b_inte : Byte);
{ sets the r, g, and b values of a palette index }
BEGIN
  Port[$3C8] := index2set;
  Port[$3C9] := r_inte;
  Port[$3C9] := g_inte;
  Port[$3C9] := b_inte;
END;

PROCEDURE BurstSetPalette(burstPalette : RGBPalette);
VAR
  burstCount : Word;
BEGIN
  Port[$3C8] := 0;
  FOR burstCount := 0 TO 767 DO Port[$3C9] := burstPalette[burstCount];
END;

PROCEDURE WaitForRetrace;
{ waits for a vertical retrace to reduce flicker }
BEGIN
  REPEAT UNTIL (Port[$3DA] AND 8) = 8;
END;

PROCEDURE ClearScr;
BEGIN
  FillChar(Mem[$A000:0000], 64000, 0);
END;

PROCEDURE FillBox(X1,Y1,x2,y2:Integer; Col:Byte);
{This draws a filled rectangle from (x1,y1) to (x2,y2)}
VAR loop,l2:Word;
BEGIN
     FOR loop:=Y1 TO y2 DO
         FOR l2:=X1 TO x2 DO
             PutPixel(l2,loop,Col);
END;

PROCEDURE screen_off;{******************************************************}

VAR regs :registers;

BEGIN
  regs.ah := $12;                    { 12 = vgahi 640 x 480 }
  regs.al := Ord(1);                 { 0 = on, 1 = off }
  regs.bl := $36;                    { Subfunction }
  Intr($10, regs);                   { Call BIOS }
END;

PROCEDURE screen_on;{*******************************************************}

VAR regs :registers;

BEGIN
  regs.ah := $12;                    { 12 = vgahi 640 x 480 }
  regs.al := Ord(0);                 { 0 = on, 1 = off }
  regs.bl := $36;                    { Subfunction }
  Intr($10, regs);                   { Call BIOS }
END;

PROCEDURE msdelay(ms : Word); Assembler;{**************************************}
ASM
  mov ax, 1000;
  mul ms;
  mov cx, dx;
  mov dx, ax;
  mov ah, $86;
  Int $15;
END;

PROCEDURE DrawBar( X1,Y1,x2,y2 :Word; Color :Byte );{***********************}

VAR Row :Word;

BEGIN
  FOR Row := Y1 TO y2 DO FillChar(Mem[$A000:(320*Row)+X1],x2-X1+1,Color);
END;

PROCEDURE Rectangle(X,y,xx,yy :Integer;Color :Byte);{***********************}

VAR
  p :Integer;

BEGIN
  FOR p :=X TO xx DO PutPixel(p,y,Color);
  FOR p :=X TO xx DO PutPixel(p,yy,Color);
  FOR p :=y TO yy DO PutPixel(X,p,Color);
  FOR p :=y TO yy DO PutPixel(xx,p,Color);
END;

PROCEDURE Rotate(cent1,cent2,angle:Integer;coord1,coord2:Real;clr:Word);
VAR coord1t, coord2t : Real;
    c1, c2 : Integer;
BEGIN
  coord1t := coord1 - cent1;
  coord2t := coord2 - cent2;
  coord1 := coord1t * Cos(angle * Pi / 180) - coord2t * Sin(angle * Pi / 180);
  coord2 := coord1t * Sin(angle * Pi / 180) + coord2t * Cos(angle * Pi / 180);
  coord1 := coord1 + cent1;
  coord2 := coord2 + cent2;
  c1 := Round(coord1);
  c2 := Round(coord2);
  PutPixel(c1,c2,clr);
END;

PROCEDURE Ellipse2(exc, eyc, ea, eb, i, clr : Integer);
VAR
  elx, ely : Integer;
  aa, aa2, bb, bb2, d, dx, dy : longint;
  X,y : Real;
BEGIN
  elx := 0; ely := eb; aa := longint(ea) * ea; aa2 := 2 * aa;
  bb := longint(eb) * eb; bb2 := 2 * bb;
  d := bb - aa * eb + aa DIV 4; dx := 0; dy := aa2 * eb;
  X := exc; y := eyc - ely;
  Rotate(exc,eyc,i,X,y,clr);
  X := exc; y := eyc + ely;
  Rotate(exc,eyc,i,X,y,clr);
  X := exc - ea; y := eyc;
  Rotate(exc,eyc,i,X,y,clr);
  X := exc + ea; y := eyc;
  Rotate(exc,eyc,i,X,y,clr);
  WHILE (dx < dy) DO BEGIN
    IF (d > 0) THEN BEGIN Dec(ely); Dec(dy, aa2); Dec(d, dy); END;
    Inc(elx); Inc(dx, bb2); Inc(d, bb + dx);
    X := exc + elx; y := eyc + ely;
    Rotate(exc,eyc,i,X,y,clr);
    X := exc - elx; y := eyc + ely;
    Rotate(exc,eyc,i,X,y,clr);
    X := exc + elx; y := eyc - ely;
    Rotate(exc,eyc,i,X,y,clr);
    X := exc - elx; y := eyc - ely;
    Rotate(exc,eyc,i,X,y,clr);
  END;
  Inc(d, (3 * (aa - bb) DIV 2 - (dx + dy)) DIV 2);
  WHILE (ely > 0) DO BEGIN
    IF (d < 0) THEN BEGIN Inc(elx); Inc(dx, bb2); Inc(d, bb + dx); END;
    Dec(ely); Dec(dy, aa2); Inc(d, aa - dy);
    X := exc + elx; y := eyc + ely;
    Rotate(exc,eyc,i,X,y,clr);
    X := exc - elx; y := eyc + ely;
    Rotate(exc,eyc,i,X,y,clr);
    X := exc + elx; y := eyc - ely;
    Rotate(exc,eyc,i,X,y,clr);
    X := exc - elx; y := eyc - ely;
    Rotate(exc,eyc,i,X,y,clr);
  END;
END;

PROCEDURE HLine(X,y,x2: Integer; Color: Byte);
BEGIN
  FOR X := X TO x2 DO PutPixel(X,y,Color);
END;

PROCEDURE VLine(X,y,y2: Integer; Color: Byte);
BEGIN
  FOR y := y TO y2 DO PutPixel(X,y,Color);
END;

PROCEDURE Circle(X,y,size: longint; Color: Byte);
VAR Xl,Yl : longint;
BEGIN
  IF size=0 THEN BEGIN
    PutPixel(X,y,Color);
    Exit;
  END;
  Xl := 0;
  Yl := size;
  size := size*size+1;
  REPEAT
    PutPixel(X+Xl,y+Yl,Color);
    PutPixel(X-Xl,y+Yl,Color);
    PutPixel(X+Xl,y-Yl,Color);
    PutPixel(X-Xl,y-Yl,Color);
    IF Xl*Xl+Yl*Yl >= size THEN Dec(Yl)
    ELSE Inc(Xl);
  UNTIL Yl = 0;
  PutPixel(X+Xl,y+Yl,Color);
  PutPixel(X-Xl,y+Yl,Color);
  PutPixel(X+Xl,y-Yl,Color);
  PutPixel(X-Xl,y-Yl,Color);
END;

PROCEDURE polygon(X1,Y1,x2,y2,x3,y3,x4,y4:Integer; c:Byte);
VAR
  xpos:ARRAY[0..199,0..1] OF Integer;
  mny,mxy,y:Integer;
  i:Word;
  s1,s2,s3,s4:shortint;
  Pos:Byte;
BEGIN
  mny:=Y1;
  IF y2<mny THEN mny:=y2;
  IF y3<mny THEN mny:=y3;
  IF y4<mny THEN mny:=y4;

  mxy:=Y1;
  IF y2>mxy THEN mxy:=y2;
  IF y3>mxy THEN mxy:=y3;
  IF y4>mxy THEN mxy:=y4;

  s1:=Byte(Y1<y2)*2-1;
  s2:=Byte(y2<y3)*2-1;
  s3:=Byte(y3<y4)*2-1;
  s4:=Byte(y4<Y1)*2-1;

  y:=Y1; Pos:=Byte(Y1<y2);
  IF Y1<>y2 THEN REPEAT
    xpos[y,Pos]:=Integer(x2-X1)*(y-Y1) DIV (y2-Y1)+X1;
    Inc(y,s1);
  UNTIL y=y2+s1 ELSE xpos[y,Pos]:=X1;
  y:=y2; Pos:=Byte(y2<y3);
  IF y2<>y3 THEN REPEAT
    xpos[y,Pos]:=Integer(x3-x2)*(y-y2) DIV (y3-y2)+x2;
    Inc(y,s2);
  UNTIL y=y3+s2 ELSE xpos[y,Pos]:=x2;
  y:=y3; Pos:=Byte(y3<y4);
  IF y3<>y4 THEN REPEAT
    xpos[y,Pos]:=Integer(x4-x3)*(y-y3) DIV (y4-y3)+x3;
    Inc(y,s3);
  UNTIL y=y4+s3 ELSE xpos[y,Pos]:=x3;
  y:=y4; Pos:=Byte(y4<Y1);
  IF y4<>Y1 THEN REPEAT
    xpos[y,Pos]:=Integer(X1-x4)*(y-y4) DIV (Y1-y4)+x4;
    Inc(y,s4);
  UNTIL y=Y1+s4 ELSE xpos[y,Pos]:=x4;
  FOR y:=mny TO mxy DO
    HLine(xpos[y,0],xpos[y,1],y,c);
END;

PROCEDURE WriteXY( X,y :Word; Color :Byte; Str :STRING );
VAR
   OldX :Word;                  { Holds Original Column.                }
   OldY :Word;
   StrPos :Byte;                { Character pos in string to write.     }
   Len    :Word;
   FontChr:Byte;                { ROM font info.                        }
   FontPos:Byte;
   BitPos :Byte;
BEGIN
  OldY := y;                    { Save Starting Row.                    }
  Len := Length(Str);           { Save the Length                       }
  FOR StrPos := 1 TO Len DO
  BEGIN                         { Loop through every character.         }
   OldX := X;                   { Save Current Column.                  }
   y := OldY;                   { Restore starting row.                 }
   FOR FontPos := 0 TO 7 DO
   BEGIN                        { Scroll through all 8 BYTES of font.   }
    FontChr := Mem[$FFA6:$E+(Ord(Str[StrPos]) SHL 3) + FontPos];
    FOR BitPos := 7 DOWNTO 0 DO
    BEGIN                       { Scroll through all 8 BITS of each BYTE.}
     IF (FontChr AND (1 SHL BitPos)) <> 0 THEN
      PutPixel(X,y,Color);      { IF bit is set then draw pixel.        }
     Inc(X);                    { point to next column.                 }
    END;
    Inc(y);                     { point to next row.                    }
    X := OldX;                  { Restore old column for next line.     }
   END;
   X := X + 8;                  { Move 9 columns ahead.                 }
  END;
END;

END.
